"""
pydtn is a module for simulating delay tolerant networks.

A simulation is made up from a Network which contains:
- A group of nodes which are responsible for forwarding packets to each other.
- A trace which defines when nodes come into contact and when they are no
  longer in contact.
- A traffic pattern which defines when a packet is created, which node is the
  source, and which node is the destination.
"""

__all__ = [
    'Network',
    'Buffer',
    'PacketInfo',

    'Node',
    'EpidemicNode',
    'FloodingNode',

    'Contact',
    'Trace',
    'CSVTrace',
    'RandomTrace',

    'Traffic',
    'RandomManyToManyTraffic',
    'RandomManyToOneTraffic',
    'RandomOneToManyTraffic',
    'RandomOneToOneTraffic',
    'RandomTraffic',
]
__version__ = '0.2'
__author__ = 'Jarrod Pas <j.pas@usask.ca>'

from collections import ChainMap, defaultdict, namedtuple, OrderedDict
import csv
from itertools import count
from random import Random
from time import time

import simpy


class Network:
    """
    Network simulation.

    Keyword Arguments:
    - nodes -- the nodes of the network in a dictionary of id to node.
    - trace -- a generator that yield Contact objects.
    - traffic -- a generator that yields Traffic objects.

    Running a simulation:
    To run the simulation call the `run` method.

    Statistics:
    There are two ways to get statistics from the simulation, via either the
    `stats` or the `stats_summary` property.
    """

    def __init__(self, nodes=None, trace=None, traffic=None):
        """Create a network."""
        self.env = simpy.Environment()

        self._stats = defaultdict(int)
        self._stats['wall-time'] = 0

        if nodes is None:
            nodes = {}
        self.nodes = nodes

        if not trace:
            trace = []
        self.trace = trace
        self.trace_done = simpy.Event(self.env)

        if not traffic:
            traffic = []
        self.traffic = traffic

        self.packets = []

        # start components in order of priority
        self.env.process(self.traffic_loop())
        self.env.process(self.trace_loop())

        for node in self.nodes.values():
            node.start(self)

    def traffic_loop(self):
        """Trigger traffic for nodes."""
        for traffic in self.traffic:
            if traffic.created > self.now:
                yield self.env.timeout(traffic.created - self.now)

            packet = Packet(self, traffic)
            self.packets.append(packet)

            packet.source.recv(packet)

    def trace_loop(self):
        """Trigger contacts for nodes."""
        for contact in self.trace:
            if contact.time > self.now:
                yield self.env.timeout(contact.time - self.now)

            node_a = self.nodes[contact.a]
            node_b = self.nodes[contact.b]

            if contact.join:
                node_a.join(node_b)
                node_b.join(node_a)
            else:
                node_a.leave(node_b)
                node_b.leave(node_a)

        self.trace_done.succeed()

    @property
    def now(self):
        """Return current simulation time."""
        return self.env.now

    def run(self, until=None):
        """
        Run the simulation.

        Keyword Arguments:
        until -- tick to run the simulation to and stop
        """
        if until is None:
            until = self.trace_done

        start = time()
        self.env.run(until=until)
        end = time()

        self._stats['wall-time'] += end - start

    @property
    def stats(self):
        """Return detailed statistics for the simulation."""
        packets = [packet.stats for packet in self.packets]
        nodes = [node.stats for node in self.nodes.values()]

        stats = {
            'sim-time': self.env.now,
            'packets': packets,
            'nodes': nodes,
        }

        stats.update(self._stats)
        return stats

    @property
    def stats_summary(self):
        """Return summarized stats for the simulation."""
        def gather(group):
            """Gather statistics from a group and return a dict of lists."""
            stats = defaultdict(list)
            for member in group:
                for stat, value in member.stats.items():
                    stats[stat].append(value)
            return dict(stats)

        packets = gather(self.packets)
        nodes = gather(self.nodes.values())

        stats = {}
        stats['sim-time'] = self.env.now
        stats['packets'] = len(self.packets)
        stats['recieved'] = len(packets['recieved'])
        stats['duplicates'] = sum(packets['recieved']) - stats['recieved']
        stats['delivery-ratio'] = stats['recieved'] / stats['packets']
        stats['broadcasts'] = sum(nodes['broadcasts'])
        stats['delivery-cost'] = stats['broadcasts'] / stats['recieved']

        stats.update(self._stats)
        return stats


PacketInfo = namedtuple('PacketInfo', [
    'source', 'destination', 'created', 'time_to_live', 'payload'
])


class Packet:
    """An item to route through the network."""

    def __init__(self, network, info):
        """Create a packet within a network, based on traffic."""
        self.network = network
        self.info = info
        self.recieved = None
        self._stats = defaultdict(int)

    @property
    def source(self):
        """Source of packet."""
        return self.network.nodes[self.info.source]

    @property
    def destination(self):
        """Destination of packet."""
        return self.network.nodes[self.info.destination]

    @property
    def created(self):
        """Time of packet creation."""
        return self.info.created

    @property
    def time_to_live(self):
        """Time to live of packet."""
        return self.info.time_to_live

    @property
    def payload(self):
        """Payload of packet."""
        return self.info.payload

    @property
    def deadline(self):
        """Return deadline of packet."""
        return self.created + self.time_to_live

    @property
    def expired(self):
        """Is the packet expired."""
        return self.network.now > self.deadline

    def sent(self, target, reason=None):
        """Send the packet from source to taget."""
        self._stats['hops'] += 1
        if reason:
            self._stats['hops-%s' % reason] += 1

        if target is self.destination:
            if self.recieved is None:
                self.recieved = 0
                self._stats['delay'] = self.network.now - self.created
            self.recieved += 1

    @property
    def stats(self):
        """Return statistcs for a packet."""
        stats = {}

        if self.recieved:
            stats['recieved'] = self.recieved

        stats.update(self._stats)
        return stats

    def __len__(self):
        """Return size of packet."""
        if isinstance(self.payload, int):
            return self.payload
        return len(self.payload)


class Buffer:
    """
    A place for a node to hold packets.

    Stores the order that packet were added to the buffer.
    Can be removed from while being iterated over.
    """

    def __init__(self, capacity=None):
        """
        Create a buffer.

        Keyword arguments:
        capacity -- the maximum number of packets to hold (default infinity).
        """
        if capacity is None:
            self.capacity = float('inf')
        else:
            self.capacity = capacity
        self.store = OrderedDict()

    @property
    def full(self):
        """Is the buffer full."""
        return len(self.store) >= self.capacity

    def add(self, packet):
        """Add a packet to the buffer."""
        if self.full:
            return False
        self.store[packet] = None
        return True

    def remove(self, packet):
        """Remove packet from the buffer."""
        if packet in self.store:
            del self.store[packet]

    def __contains__(self, packet):
        """Is the packet in the buffer."""
        return packet in self.store

    def __iter__(self):
        """
        Return an iterator for packets in the buffer.

        Allows for removing items during iteration.
        """
        return iter(list(self.store.keys()))


class Node:
    """
    Basic implementation of a node.

    Forwards directly to destination node.
    """

    class SendFailed(Exception):
        """Raised when a send fails."""

    def __init__(self, **options):
        """Create a node."""
        self.network = None

        self.buffer = Buffer(capacity=options.get('buffer_capacity', None))
        self._neighbours = set()

        self.options = ChainMap(options, {
            'bandwidth': float('inf'),
            'tick_rate': 1,
        })

        self._stats = defaultdict(int)
        self._stats['broadcasts'] = 0

    def start(self, network):
        """
        Start event loop.

        If it has already been started do nothing.
        """
        if self.network is not None:
            return

        self.network = network
        self.network.env.process(self.tick())

    def tick(self):
        """Thread of execution for a node."""
        tick = self.options['tick_rate']
        while True:
            start = self.network.now
            yield from self.forward_all()
            delay = tick - (self.network.now - start) % tick
            yield self.network.env.timeout(delay)

    def join(self, node):
        """
        Approach the neighbouhood of this node.

        This is an idempotent operation.
        """
        if node not in self._neighbours:
            self._neighbours.add(node)

    def leave(self, node):
        """
        Leave the neighbouhood of this node.

        This is an idempotent operation.
        """
        if node in self._neighbours:
            self._neighbours.remove(node)

    @property
    def neighbours(self):
        """Return nodes within transmission range."""
        return frozenset(self._neighbours)

    def forward(self, packet):
        """Forward packet directly to their destination if possible."""
        if packet.destination in self.neighbours:
            return {packet.destination: 'direct'}
        # forward to nobody
        return {}

    def forward_all(self):
        """Try to forward all packets."""
        env = self.network.env

        for packet in self.buffer:
            if packet.expired:
                self.packet_expiry(packet)
                continue

            forwards = self.forward(packet)
            if forwards:
                delay = len(packet) / self.options['bandwidth']
                yield env.timeout(delay)
                self._stats['broadcasts'] += 1

            for target, reason in forwards.items():
                try:
                    self.send(packet, target, reason)
                    self.send_success(packet, target)
                except Node.SendFailed:
                    self.send_failure(packet, target)

    def send(self, packet, target, reason=None):
        """
        Send a packet to another node.

        Keyword Arguments:
        reason -- why the packet was sent (optional)
        """
        if target in self.neighbours:
            target.recv(packet, source=target)
            packet.sent(target, reason=reason)
        else:
            raise Node.SendFailed('nodes are not neighbours')

    def send_success(self, packet, target):
        """
        Call when a send succeeds.

        Removes the sent packet from the buffer.
        """
        # pylint: disable=unused-argument
        self.buffer.remove(packet)

    def send_failure(self, packet, target):
        """Call when a send fails."""
        # pylint: disable=unused-argument
        pass

    def packet_expiry(self, packet):
        """Call when a packet expires."""
        self.buffer.remove(packet)

    def recv(self, packet, source=None):
        """
        Recieve a packet.

        If the packet has reached it's destination notes the time then
        """
        # pylint: disable=unused-argument
        if packet.destination is self:
            return

        if not self.buffer.add(packet):
            raise Node.SendFailed('buffer full')

    @property
    def stats(self):
        """Return node statistics."""
        return dict(self._stats)

    def __repr__(self):
        """Return representation."""
        return '%s(**%r)' % (
            self.__class__.__name__,
            self.options
        )


class EpidemicNode(Node):
    """Node which forwards epidemically."""

    def __init__(self, **options):
        """Create an epidemic node."""
        super().__init__(**options)
        self.__sent = defaultdict(set)

    def forward(self, packet):
        """
        Forward based on the epidemic heuristic.

        Arguments:
        packet -- packet to be forwarded.

        Epidemic Heuristic:
        - Forward the packet to all neighbours that I have not forwarded
          to so far.
        """
        forward = {
            neighbour: 'epidemic'
            for neighbour in self.neighbours
            if neighbour not in self.__sent[packet]
        }

        return forward

    def send_success(self, packet, target):
        """
        Call when a send succeeds.

        Arguments:
        packet -- packet that was sent successfully.
        target -- the node that packet was sent to.

        Adds target to packet tracking.
        """
        super().send_success(packet, target)
        self.__sent[packet].add(target)

    def packet_expiry(self, packet):
        """
        Call when a packet expires.

        Removes packet from tracking.
        """
        super().packet_expiry(packet)
        if packet in self.__sent:
            del self.__sent[packet]


class FloodingNode(Node):
    """Node which forwards through flooding."""

    def forward(self, packet):
        """Forward to all neighbour nodes."""
        forward = {
            neighbour: 'flood'
            for neighbour in self.neighbours
        }

        return forward

    def send_success(self, packet, target):
        """
        Call when send suecceeds.

        Do nothing, overrides default of removing from buffer.
        """
        pass


Contact = namedtuple('Contact', ['time', 'a', 'b', 'join'])


class Trace:
    """Contact trace base class."""

    def __init__(self, nodes):
        """Create a contact trace."""
        self._nodes = nodes

    @property
    def nodes(self):
        """Return number of nodes in trace."""
        return self._nodes

    def __iter__(self):
        """Yield no contacts."""
        yield

    @staticmethod
    def create_contact(now, node_a, node_b, join):
        """Create a contact."""
        return Contact(time=now, a=node_a, b=node_b, join=join)


class CSVTrace(Trace):
    """
    Contact trace from csv file.

    The first row is used as headers.

    The second row is the following information about the trace:
    [ int(duration), int(node_count), ignored, ignored ]

    The csv uses the following columns:
    [ int(time), int(node_a), int(node_b), int(join) ]
    time -- simulation time that the contact even happens
    node_{a,b} -- nodes involved in contact
    join -- whether the contact is going up (1) or going down (0)
    """

    def __init__(self, path):
        """Create a csv trace generator."""
        with open(path) as csv_file:
            csv_file = csv.reader(csv_file)
            # skip header
            next(csv_file)
            # get node count from first row of csv. it is additional stats
            # about the trace
            # duration, placeholder, placehold, node_count
            nodes = int(next(csv_file)[1])

        super().__init__(nodes)
        self.path = path

    def __iter__(self):
        """Yield contacts from csv file."""
        with open(self.path) as csv_file:
            csv_file = csv.reader(csv_file)
            # skip header and first row
            next(csv_file)
            next(csv_file)
            for row in csv_file:
                yield self.create_contact(*map(int, row))


class RandomTrace(Trace):
    """Contact trace from complete randomness."""

    def __init__(self, nodes, seed=None, step=1):
        """Create a random trace generator."""
        super().__init__(len(nodes))
        self._node_list = list(nodes)
        self.step = step
        self.seed = seed

    def __iter__(self):
        """Yield random trace."""
        random = Random(self.seed)

        for now in count(step=self.step):
            join = bool(random.getrandbits(1))
            node_a, node_b = random.sample(self._node_list, 2)
            yield self.create_contact(now, node_a, node_b, join)


class Traffic:
    """Base traffic generator."""

    # pylint: disable=too-few-public-methods

    def __init__(self, **options):
        """Create traffic generator."""
        self.time_to_live = options.get('time_to_live', float('inf'))
        self.payload = options.get('payload', 0)

    def __iter__(self):
        """Yield no traffic."""
        yield

    def create_packet(self, now, source, destination):
        """Create a packet info for the simulation."""
        return PacketInfo(source=source,
                          destination=destination,
                          created=now,
                          time_to_live=self.time_to_live,
                          payload=self.payload)


class RandomManyToManyTraffic(Traffic):
    """Generate random traffic between nodes."""

    def __init__(self, sources, destinations, **options):
        """
        Create a random traffic generator.

        Option Arguments:
        seed -- seed to use for the random number generator (default None)
        start -- which time step the first packet is create (default 0)
        step -- time between traffic generation (default 1)
        traffic_per_step -- traffic to generate during each step (default 1)
        time_to_live -- packet time to live (default infinite)
        payload -- packet payload (default 0)
        """
        super().__init__(**options)

        self.sources = list(sources)
        self.destinations = list(destinations)

        if len(self.sources) == 1 and self.sources == self.destinations:
            raise ValueError('only source is the same as only destination')

        self.seed = options.get('seed', None)
        self.start = options.get('start', 0)
        self.step = options.get('step', 1)
        self.traffic_per_step = options.get('traffic_per_step', 1)

    def random_pair(self, random):
        """Return random pair of nodes."""
        source = destination = random.choice(self.sources)
        while destination == source:
            destination = random.choice(self.destinations)
        return source, destination

    def __iter__(self):
        """Yield infinite random traffic."""
        random = Random(self.seed)

        for now in count(start=self.start, step=self.step):
            for _ in range(self.traffic_per_step):
                source, destination = self.random_pair(random)
                yield self.create_packet(now, source, destination)


class RandomManyToOneTraffic(RandomManyToManyTraffic):
    """Generate random traffic from any source to a single node."""

    def __init__(self, sources, destination, **options):
        """Create many to one traffic generator."""
        super().__init__(sources, [destination], **options)


class RandomOneToManyTraffic(RandomManyToManyTraffic):
    """Generate random traffic from a single node to any destination."""

    def __init__(self, source, destinations, **options):
        """Create one to many traffic generator."""
        super().__init__([source], destinations, **options)


class RandomOneToOneTraffic(RandomManyToManyTraffic):
    """Generate random traffic from a single node to a single node."""

    def __init__(self, source, destination, **options):
        """Create one to one traffic generator."""
        super().__init__([source], [destination], **options)


class RandomTraffic(RandomManyToManyTraffic):
    """Generate random traffic between any node a group."""

    def __init__(self, nodes, **options):
        """Create random traffic generator."""
        super().__init__(nodes, nodes, **options)
